/*

mod_security, http://www.webkreator.com/mod_security/
Ivan Ristic, ivanr@webkreator.com

Copyright (c) 2002, WebKreator http://www.webkreator.com/
All rights reserved.

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

  * Redistributions of source code must retain the above
    copyright notice, this list of conditions and the following
    disclaimer.
      
  * Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
      
  * Neither the name of the WebKreator nor the names of its
    contributors may be used to endorse or promote products derived
    from this software without specific prior written permission.
            
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
            
*/            

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_core.h"
#include "http_main.h"
#include "http_log.h"

module MODULE_VAR_EXPORT security_module;

#define ACTION_NONE             0
#define ACTION_DENY             1

#define VAR_ACTION_ALLOW        1
#define VAR_ACTION_DENY         0

#define VAR_UNKNOWN             0
#define VAR_CUSTOM 		        10000
#define VAR_HEADER 		        10001
#define VAR_ENV	   		        10002
#define VAR_ARGS   		        10003
#define VAR_POST_PAYLOAD	    10004

#define VAR_REMOTE_ADDR		    1
#define VAR_REMOTE_HOST		    2
#define VAR_REMOTE_USER		    3   
#define VAR_REMOTE_IDENT	    4
#define VAR_REQUEST_METHOD	    5
#define VAR_SCRIPT_FILENAME	    6
#define VAR_PATH_INFO		    7
#define VAR_QUERY_STRING	    8
#define VAR_AUTH_TYPE		    9
#define VAR_DOCUMENT_ROOT	    10
#define VAR_SERVER_ADMIN	    11
#define VAR_SERVER_NAME		    12
#define VAR_SERVER_ADDR		    13
#define VAR_SERVER_PORT		    14
#define VAR_SERVER_PROTOCOL	    15
#define VAR_SERVER_SOFTWARE	    16
#define VAR_TIME_YEAR		    17
#define VAR_TIME_MON		    18
#define VAR_TIME_DAY		    19
#define VAR_TIME_HOUR		    20
#define VAR_TIME_MIN		    21
#define VAR_TIME_SEC		    22
#define VAR_TIME_WDAY		    23
#define VAR_TIME		        24
#define VAR_API_VERSION		    25
#define VAR_THE_REQUEST		    26
#define VAR_REQUEST_URI		    27
#define VAR_REQUEST_FILENAME	28
#define VAR_IS_SUBREQ		    29

char *all_variables[] = {
    "REMOTE_ADDR",
    "REMOTE_HOST",
    "REMOTE_USER",
    "REMOTE_IDENT",
    "REQUEST_METHOD",
    "SCRIPT_FILENAME",
    "PATH_INFO",
    "QUERY_STRING",
    "AUTH_TYPE",
    "DOCUMENT_ROOT",
    "SERVER_ADMIN",
    "SERVER_NAME",
    "SERVER_ADDR",
    "SERVER_PORT",
    "SERVER_PROTOCOL",
    "SERVER_SOFTWARE",
    "TIME_YEAR",
    "TIME_MON",
    "TIME_DAY",
    "TIME_HOUR",
    "TIME_MIN",
    "TIME_SEC",
    "TIME_WDAY",
    "TIME",
    "API_VERSION",
    "THE_REQUEST",
    "REQUEST_URI",
    "REQUEST_FILENAME",
    "IS_SUBREQ",
    NULL
};

typedef struct {
    char *name;
    int type;
    int action;
} variable;

typedef struct {
    char *buffer;
    char *sofar;
    long length;
    long remaining;
} request_body;

typedef struct {
    int log;
    int action;
    int status;
    regex_t *regex;
    int is_default_action_used;
    int is_selective;
    int is_negative;
    int is_allow;
    int requires_parsed_args;
    array_header *variables;
} signature;

typedef struct {
    int filter_engine;
    int scan_post;
    signature action;
    array_header *signatures;
    char *path;
    int auditlog_flag;
    char *auditlog_name;
    int auditlog_fd;
    int filter_debug_level;
} sec_dir_config;

typedef struct {
    int filter_engine;
    int scan_post;
    signature action;
    array_header *signatures;
    char *path;
    int auditlog_flag;
    char *auditlog_name;
    int auditlog_fd;
    int filter_debug_level;
} sec_srv_config;

const char * get_variable (request_rec *r, variable *v, table *parsed_args) {
    time_t tc;
    struct tm *tm;
    static char resultbuf[254] = "";
    const char *result = NULL;

    switch (v->type) {

    case VAR_CUSTOM:

    	if (parsed_args != NULL) {
	        result = ap_table_get(parsed_args, v->name);
    	}
	    else {
    	    // log error
	    }

    	if (result == NULL) {
	        *resultbuf = 0;
	        result = resultbuf;
    	}
	    break;

    case VAR_HEADER:
    	result = ap_table_get(r->headers_in, v->name);
	    break;

    case VAR_ENV:
    	result = ap_table_get(r->notes, v->name);

	    if (result == NULL) {
	        result = ap_table_get(r->subprocess_env, v->name);
    	}

	    if (result == NULL) {
	        result = getenv(v->name);
    	}
    	break;

    case VAR_ARGS:
    	strcpy(resultbuf, "(internal error)");
	    result = resultbuf;
    	break;

    case VAR_REMOTE_ADDR:
    	result = r->connection->remote_ip;
	    break;

    case VAR_REMOTE_HOST:
    	result = ap_get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME);
        break;

    case VAR_REMOTE_USER:
        result = r->connection->user;
	    break;

    case VAR_REMOTE_IDENT:
        result = ap_get_remote_logname(r);
        break;

    case VAR_REQUEST_METHOD:
        result = r->method;
        break;

    case VAR_REQUEST_URI:
        result = r->uri;
	    break;

    case VAR_AUTH_TYPE:
    	result = r->connection->ap_auth_type;
	    break;

    case VAR_IS_SUBREQ:
    	result = (r->main != NULL ? "true" : "false");
	    break;

    case VAR_DOCUMENT_ROOT:
    	result = ap_document_root(r);
	    break;

    case VAR_SERVER_ADMIN:
    	result = r->server->server_admin;
	    break;

    case VAR_SERVER_NAME:
    	result = ap_get_server_name(r);
	    break;

    case VAR_SERVER_ADDR:
    	result = r->connection->local_ip;
	    break;

    case VAR_SERVER_PORT:
    	ap_snprintf(resultbuf, sizeof(resultbuf), "%u", ap_get_server_port(r));
    	result = resultbuf;
	    break;

    case VAR_SERVER_PROTOCOL:
    	result = r->protocol;
	    break;

    case VAR_SERVER_SOFTWARE:
    	result = ap_get_server_version ();
	    break;

    case VAR_API_VERSION:
        ap_snprintf(resultbuf, sizeof(resultbuf), "%d:%d", MODULE_MAGIC_NUMBER_MAJOR, MODULE_MAGIC_NUMBER_MINOR);
        result = resultbuf;

    case VAR_TIME_YEAR:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf (resultbuf, sizeof(resultbuf), "%02d%02d", (tm->tm_year / 100) + 19, tm->tm_year % 100);
        result = resultbuf;
        break;

    case VAR_TIME:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf (resultbuf, sizeof (resultbuf), "%02d%02d%02d%02d%02d%02d%02d",
		     (tm->tm_year / 100) + 19, (tm->tm_year % 100),
		     tm->tm_mon + 1, tm->tm_mday, tm->tm_hour, tm->tm_min,
		     tm->tm_sec);
	     
	    result = resultbuf;
    	break;

    case VAR_TIME_WDAY:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf(resultbuf, sizeof(resultbuf), "%d", tm->tm_wday);
        result = resultbuf;
        break;

    case VAR_TIME_SEC:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf(resultbuf, sizeof(resultbuf), "%02d", tm->tm_sec);
        result = resultbuf;
        break;

    case VAR_TIME_MIN:
        tc = time (NULL);	
        tm = localtime (&tc);
        ap_snprintf(resultbuf, sizeof(resultbuf), "%02d", tm->tm_min);
        result = resultbuf;
        break;

    case VAR_TIME_HOUR:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf(resultbuf, sizeof(resultbuf), "%02d", tm->tm_hour);
        result = resultbuf;
        break;

    case VAR_TIME_MON:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf (resultbuf, sizeof(resultbuf), "%02d", tm->tm_mon + 1);
        result = resultbuf;
        break;

    case VAR_TIME_DAY:
        tc = time (NULL);
        tm = localtime (&tc);
        ap_snprintf(resultbuf, sizeof(resultbuf), "%02d", tm->tm_mday);
        result = resultbuf;
        break;

    case VAR_SCRIPT_FILENAME:
    case VAR_REQUEST_FILENAME:
        result = r->filename;
        break;

    case VAR_PATH_INFO:
        result = r->path_info;
        break;

    case VAR_THE_REQUEST:
        result = r->the_request;
        break;

    case VAR_QUERY_STRING:
        result = r->args;
        break;
    }

    if (result == NULL) {
        *resultbuf = 0;
        result = resultbuf;
    }

    return result;
}

request_rec *find_last_request(request_rec *r) {
    while (r->next != NULL) r = r->next;
    return r;
}

static void *sec_create_srv_config(pool *p, server_rec *s) {
    sec_srv_config *scfg = ap_palloc(p, sizeof(sec_srv_config));

    scfg->filter_engine = 0;
    scfg->scan_post = 0;

    scfg->action.log = 1;
    scfg->action.action = ACTION_DENY;
    scfg->action.status = HTTP_FORBIDDEN;

    scfg->signatures = ap_make_array (p, 10, sizeof (signature *));

    scfg->path = ap_pstrdup (p, ":server");

    scfg->auditlog_flag = 0;
    scfg->auditlog_name = NULL;
    scfg->auditlog_fd = -1;
    
    scfg->filter_debug_level = 0;

    return scfg;
}

static void *sec_create_dir_config(pool *p, char *path) {
    sec_dir_config *dcfg = (sec_dir_config *)ap_pcalloc (p, sizeof(*dcfg));

    dcfg->filter_engine = -1;
    dcfg->scan_post = -1;

    dcfg->action.log = 1;
    dcfg->action.action = ACTION_DENY;
    dcfg->action.status = HTTP_FORBIDDEN;

    dcfg->signatures = ap_make_array(p, 10, sizeof(signature *));

    if (path == NULL) {
        dcfg->path = ap_pstrdup (p, ":null");
    }
    else {
        dcfg->path = ap_pstrdup (p, path);
    }

    dcfg->auditlog_flag = -1;
    dcfg->auditlog_name = NULL;
    dcfg->auditlog_fd = -1;
    
    dcfg->filter_debug_level = 0;

    return dcfg;
}

static void *sec_merge_dir_config (struct pool *p, void *_parent, void *_child) {
    sec_dir_config *parent = (sec_dir_config *)_parent;
    sec_dir_config *child = (sec_dir_config *)_child;

    // merge child & parent into new

    sec_dir_config *new = (sec_dir_config *)ap_pcalloc (p, sizeof(*new));

    memcpy(new, child, sizeof(*child));

    new->signatures = ap_copy_array(p, child->signatures);
    ap_array_cat(new->signatures, parent->signatures);

    if (child->scan_post == -1) {
    	new->scan_post = parent->scan_post;
    }
    else {
    	new->scan_post = child->scan_post;
    }

    if (child->filter_engine == -1) {
    	new->filter_engine = parent->filter_engine;
    }
    else {
    	new->filter_engine = child->filter_engine;
    }

    if (child->auditlog_fd == -1) {
	    new->auditlog_fd = parent->auditlog_fd;
    	new->auditlog_name = parent->auditlog_name;
    }
    else {
	    new->auditlog_fd = child->auditlog_fd;
    	new->auditlog_name = child->auditlog_name;
    }

    return new;
}

char x2c (char *what) {
    register char digit;

    digit = (what[0] >= 'A' ? ((what[0] & 0xdf) - 'A') + 10 : (what[0] - '0'));
    digit *= 16;
    digit += (what[1] >= 'A' ? ((what[1] & 0xdf) - 'A') + 10 : (what[1] - '0'));
    
    return (digit);
}

char *normalise_uri_inplace(request_rec *r, char *uri) {
    char *p_read = uri;
    char *p_write = uri;
    char *p_slash = NULL;
    char c = 0;

    while (*p_read != 0) {

	switch (*p_read) {

    	case '/':
    	    if (p_slash == NULL) {
        		p_slash = p_read;
                *p_write++ = *p_read++;
    	    }
	        else {
		        p_read++;
    	    }
	        break;

    	case '%':
	        p_read++;
	        c = x2c (p_read);
    	    p_read += 2;
    
	            if (c < 32) {
                    ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Invalid request", r->uri);
                    return NULL;
                }

	        *p_write++ = c;
    	    break;

	    default:
    	    p_slash = NULL;
	        *p_write++ = *p_read++;
    	    break;
	    }

    }

    *p_write = 0;

    return uri;
}

char *normalise_uri(request_rec *r, char *_uri) {
    char *uri = ap_pstrdup (r->pool, _uri);
    return normalise_uri_inplace(r, uri);
}



static int read_post_payload(request_rec *r, char **p) {
    request_body *rb = ap_palloc(r->pool, sizeof(request_body));
    sec_dir_config *dcfg = (sec_dir_config *)ap_get_module_config(r->per_dir_config, &security_module);

    if (dcfg->filter_engine != 1) return DECLINED;
    if (dcfg->scan_post != 1) return DECLINED;

    // is this a POST request?
    // if it is, we will handle it ourselves
    if (r->method_number == M_POST) {
    	char buf[100];
	    char *payload, *payload2, *t;
    	int i, rc, payload_size = 0;

    	if ((rc = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR)) != OK) {
	        return rc;
    	}

	    *p = payload = t = ap_palloc (r->pool, r->remaining + 1);

    	ap_hard_timeout ("mod_security: receive request data", r);

	    if (ap_should_client_block (r)) {

	        while ((i = ap_get_client_block (r, t, r->remaining)) > 0) {
        		payload_size += i;
	        	t += i;
		        ap_reset_timeout (r);
    	    }
	    }

    	ap_kill_timeout (r);

	    payload[payload_size] = 0;

    	rb->buffer = payload;
	    rb->sofar = payload;
    	rb->length = payload_size;
	    rb->remaining = payload_size;

    	ap_table_setn (r->notes, ap_pstrdup (r->pool, "mod_security-note"), (char *) rb);
    }

    return OK;
}

char * unescape_url (char *url) {
    register int x, y;

    for (x = 0, y = 0; url[y]; ++x, ++y) {
    	if ((url[x] = url[y]) == '%') {
	        url[x] = x2c (&url[y + 1]);
	        y += 2;
    	}
    }
    
    url[x] = '\0';
    return url;
}

void parse_arguments (char *s, table * parsed_args) {
    int inputlength;
    char *buf;
    int i, j, status;
    char *value;

    if (s == NULL) return;

    inputlength = strlen (s);

    buf = (char *) malloc (inputlength + 1);
    if (buf == NULL) {
    	// log error
	    return;
    }

    i = 0;
    j = 0;
    status = 0;
    while (i < inputlength) {
    
    	while ((s[i] != '=') && (s[i] != '&') && (i < inputlength)) {

	        if (s[i] == '+') buf[j] = 32;
	        else buf[j] = s[i];

            j++;
            i++;
    	}

	    buf[j++] = '\0';

	    if (status == 0) {
	        status = 1;
	        unescape_url (buf);
	        value = &buf[j];
        }
        else {
            status = 0;
            unescape_url (value);

            ap_table_add (parsed_args, buf, value);

            j = 0;
        }
        
        i++;
    }

    if (status == 1) {
    	ap_table_add (parsed_args, buf, "");
    }

    free (buf);
    return;
}


static int sec_check_all_signatures (request_rec * r) {
    sec_dir_config *dcfg = (sec_dir_config *) ap_get_module_config (r->per_dir_config, &security_module);
    signature **signatures;
    char *_post_payload = NULL;
    table *parsed_args = NULL;
    int i, j, rs;
    char *_the_request = NULL;
    
    _the_request = normalise_uri(r, r->the_request);
    
    if (dcfg->scan_post) {
        rs = read_post_payload (r, &_post_payload);
        if (rs != OK) return rs;

        if (_post_payload != NULL) {
            // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: post payload read.");
            _post_payload = normalise_uri(r, _post_payload);
            
            if (_post_payload == NULL) {
                return dcfg->action.status;
	        }
	    }
    }


    signatures = (signature **) dcfg->signatures->elts;
    for (i = 0; i < dcfg->signatures->nelts; i++) {

        if (signatures[i]->is_selective == 0) {
            // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: non selective filter");

            // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: non selective filter against [%s]", r->the_request);
            rs = check_sig_against_string (r, signatures[i], &dcfg->action, _the_request);
	        if (rs != OK) return rs;

	        if (_post_payload != NULL) {
                // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: non selective filter against [%s]", _post_payload);
		        rs = check_sig_against_string (r, signatures[i], &dcfg->action, _post_payload);
	            if (rs != OK) return rs;
            }

        }
	    else {
	        variable **variables;
	        const char *v;

	        if ((signatures[i]->requires_parsed_args) && (parsed_args == NULL)) {
                parsed_args = ap_make_table (r->pool, 10);

        		// parse query string
        		parse_arguments (r->args, parsed_args);

    	    	// parse post payload
	    	    parse_arguments (_post_payload, parsed_args);
    	    }

	        if (signatures[i]->is_negative == 0) {

		        // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: selective filter, non-negative going into filters");

        		variables = (variable **) signatures[i]->variables->elts;
        		for (j = 0; j < signatures[i]->variables->nelts; j++) {

		            // ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r, "mod_security: processing variable [%s]", variables[j]->name);

        		    if (variables[j]->type == VAR_ARGS) {

    		        	// ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: special processing for VAR_ARGS");

	            		variables[j]->type = VAR_QUERY_STRING;
            			v = get_variable (r, variables[j], parsed_args);
            			variables[j]->type = VAR_ARGS;
            
			            if (v != NULL) {
			                rs = check_sig_against_string (r, signatures [i], &dcfg->action, v);
			                if (rs != OK) return rs;
		            	}

            			variables[j]->type = VAR_POST_PAYLOAD;
            			v = get_variable (r, variables[j], parsed_args);
            			variables[j]->type = VAR_ARGS;

			            if (v != NULL) {
			                rs = check_sig_against_string (r, signatures[i], &dcfg->action, v);
                            if (rs != OK) return rs;
			            }
		            }
		            else {

            			v = get_variable (r, variables[j], parsed_args);
			            if (v != NULL) {
			                rs = check_sig_against_string (r, signatures[i], &dcfg->action, v);
			                if (rs != OK) return rs;
            			}
			            else {
	            		    ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Variable not found [%s]", variables[j]->name);
			            }
		            }
		        }
	        }
	        else {
		        int k;
		        table_entry *te;
		        array_header *arr;
		        char *v = NULL;
		        int len = 0;

	        	if (r->args != NULL) {
		            len = strlen(r->args);
		        }

        		if (_post_payload != NULL) {
		            len += strlen(_post_payload);
        		}

		        if (len != 0) {
		            v = (char *) malloc (len + 1);
        		    if (v == NULL) {
            			// error
            			return DECLINED;
        		    }

		            *v = 0;

        		    // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: selective filter, negative starting");

        		    if (parsed_args == NULL) {
			            ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: arguments are not parsed, internal error");
			            return DECLINED;
	        	    }


        		    variables = (variable **) signatures[i]->variables->elts;
        		    for (j = 0; j < signatures[i]->variables->nelts; j++) {

		            	if ((variables[j]->type == VAR_CUSTOM) && (variables[j]->action == VAR_ACTION_ALLOW)) {
			                ap_table_unset (parsed_args, variables[j]->name);
            			}
		            }

        		    arr = ap_table_elts (parsed_args);

        		    // then filter the remaining variables
        		    te = (table_entry *) arr->elts;
        		    for (k = 0; k < arr->nelts; k++) {

		            	if (*v != 0) {
            			    strcat (v, "&");
			            }

        			strcat (v, te[k].key);
		    	    strcat (v, "=");
    			    strcat (v, te[k].val);
	    	    }

		        // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: [%s]", v);

    		    // finally, make the darn check
	    	    rs = check_sig_against_string (r, signatures[i], &dcfg->action, v);
    		    free (v);
	    	    if (rs != OK) return rs;
    		}
	    }

	}

    }

    return DECLINED;
}

int check_sig_against_string (request_rec * r, signature * _sig, signature * default_action, char *s) {

    int regex_result = regexec (_sig->regex, s, 0, NULL, 0);

    if ((regex_result == 0)&&(_sig->is_allow == 0) || (regex_result == 1)&&(_sig->is_allow == 1)) {

    	signature *sig = _sig;

	    if (sig->is_default_action_used == 1) sig = default_action;

    	if (sig->action == ACTION_DENY) {

	        if (sig->log) {
    	    	ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Access denied to signature [%s] with code [%i]", s, sig->status);
	        }
    
	        return sig->status;

	    }
    	else {
	        if (sig->log) {
    		    ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Warning for signature [%s]", s);
    	    }
	    }
    }

    return OK;
}


void parse_action (char *p2, signature * sig) {
    char *t = strdup (p2);
    char *p = strtok (t, ",");

    while (p != NULL) {
    	if (strcmp (p, "log") == 0) {
	        sig->log = 1;
    	}
	    else if (strcmp (p, "nolog") == 0) {
	        sig->log = 0;
    	}
	    else if (strncmp (p, "status", 6) == 0) {
	        if (strlen (p) > 8) {
    		sig->status = atoi (p + 7);
	        }
    	}
	    else if (strcmp (p, "deny") == 0) {
	        sig->action = ACTION_DENY;
    	}
	    else if (strcmp (p, "pass") == 0) {
	        sig->action = ACTION_NONE;
    	}

	    p = strtok (NULL, ",");
    }

    free (t);
}

static const char * cmd_filter_engine (cmd_parms * cmd, sec_dir_config * dcfg, int bool) {
    dcfg->filter_engine = bool;
    return NULL;
}

static const char * cmd_audit_engine (cmd_parms * cmd, sec_dir_config * dcfg, int bool) {
    dcfg->auditlog_flag = bool;
    return NULL;
}

static const char * cmd_audit_log (cmd_parms * cmd, sec_dir_config * dcfg, char *p1) {
    char *logname;

    // todo: open log files in the server init phase, instead of here

    dcfg->auditlog_name = ap_server_root_relative (cmd->pool, p1);
    
    dcfg->auditlog_fd = ap_popenf (cmd->pool, dcfg->auditlog_name,
		   O_WRONLY | O_APPEND | O_CREAT,
		   S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);

    if (dcfg->auditlog_fd < 0) {
    	return "mod_security: Failed to open the audit log file.";
    }

    return NULL;
}

static const char * cmd_scan_post (cmd_parms * cmd, sec_dir_config * dcfg, int bool) {
    dcfg->scan_post = bool;
    return NULL;
}

static const char * cmd_default_action (cmd_parms * cmd, sec_dir_config * dcfg, char *p1) {
    parse_action (p1, &dcfg->action);
    return NULL;
}

static const char * cmd_filter(cmd_parms * cmd, sec_dir_config * dcfg, char *p1, char *p2) {
    signature *sig;

    sig = ap_palloc (cmd->pool, sizeof (signature));
    memset (sig, 0, sizeof (signature));
    
    // p1 is the regular expression string
    if (p1[0] == '!') {
        sig->is_allow = 1;
        sig->regex = ap_pregcomp(cmd->pool, p1 + 1, REG_EXTENDED | REG_ICASE | REG_NOSUB);
    } else {
        sig->regex = ap_pregcomp(cmd->pool, p1, REG_EXTENDED | REG_ICASE | REG_NOSUB);
    }
    

    // p2 is an optional action string
    if (p2 != NULL) {
        sig->is_default_action_used = 0;
        parse_action(p2, sig);
    }
    else {
    	sig->is_default_action_used = 1;
    }

    // add the signature to the list of all signatures
    *(signature **)ap_push_array(dcfg->signatures) = sig;

    return NULL;
}

static const char *cmd_filter_debug_level(cmd_parms *cmd, sec_dir_config *dcfg, char *p1) {
	dcfg->filter_debug_level = atoi(p1);
	return NULL;
}

static const char *cmd_filter_selective(cmd_parms *cmd, sec_dir_config *dcfg, char *p1, char *p2, char *p3) {

    char *p;
    char *t;
    signature *sig;
    
    
    // initialise the structure first    
    
    sig = ap_palloc(cmd->pool, sizeof (signature));

    memset (sig, 0, sizeof (signature));

	sig->is_allow = 0;
    sig->is_selective = 1;
    sig->is_negative = 0;
    sig->requires_parsed_args = 0;
    sig->variables = ap_make_array(cmd->pool, 10, sizeof (variable *));
    
    if (p2[0] == '!') {
        sig->is_allow = 1;
        sig->regex = ap_pregcomp(cmd->pool, p2 + 1, REG_EXTENDED | REG_ICASE | REG_NOSUB);
    } else {
        sig->regex = ap_pregcomp(cmd->pool, p2, REG_EXTENDED | REG_ICASE | REG_NOSUB);
    }
    
    
    

    // split parameter 1 apart and extract variable names
    
    p = strdup(p1);
    t = strtok (p, "|");
    while (t != NULL) {

    	// add the token to the list
    	variable *v = (variable *) ap_palloc(cmd->pool, sizeof(variable));
	    char *x = t;

    	v->type = VAR_UNKNOWN;

        // when ! is the first character in the variable
        // name that means that the restrictions need to be
        // relaxed for that variable (within the filter scope)
    	if (t[0] == '!') {
    	    v->action = VAR_ACTION_ALLOW;
    	    sig->is_negative = 1;
	        sig->requires_parsed_args = 1;
    	    x++;
    	}
	    else {
    	    v->action = VAR_ACTION_DENY;
	    }

        // arguments
        if (strncmp (x, "ARG_", 4) == 0) {
	        v->type = VAR_CUSTOM;
	        v->name = ap_pstrdup(cmd->pool, x + 4);
        	sig->requires_parsed_args = 1;
        }
        // HTTP headers
        else if (strncmp (x, "HTTP_", 5) == 0) {
            char *px;

    	    v->type = VAR_HEADER;
	        v->name = ap_pstrdup (cmd->pool, x + 5);
    
      	    // replace all "_" with "-"
            px = v->name;
            while (*px != 0) {
                if (*px == '_') *px = '-';
        	        px++;
            }

    	}    
        // environment variables
        else if (strncmp (x, "ENV_", 4) == 0) {
            v->type = VAR_ENV;
            v->name = ap_pstrdup(cmd->pool, x + 4);
    	}
        // all arguments
        else if (strcmp (x, "ARGS") == 0) {
            v->type = VAR_ARGS;
	        v->name = ap_pstrdup(cmd->pool, x);
    	}
        // everything else
        else {
	        char **vl = all_variables;
	        int i = 1;

      	    while (*vl != NULL) {
       	    	if (strcmp(*vl, x) == 0) {
	   	            v->type = i;
	                v->name = ap_pstrdup(cmd->pool, x);
       	    	    break;
   	    	    }

   		        i++;
       	    	vl++;
	        }
        }

    	if (v->type == VAR_UNKNOWN) {
	    	v->name = ap_pstrdup(cmd->pool, "UKNOWN");
            // todo: throw error here
	    }

        *(variable **) ap_push_array(sig->variables) = v;

        // and proceed to the next token
        t = strtok (NULL, "|");
    }

    free(p);


    // is default action parameter present?
    if (p3 != NULL) {
        sig->is_default_action_used = 0;
        parse_action (p3, sig);
    }
    else {
        sig->is_default_action_used = 1;
    }


    // add the signature to the list of all signatures
    *(signature **) ap_push_array (dcfg->signatures) = sig;

    return NULL;
}


static const command_rec sec_cmds[] = {

     {
         "SecFilter",
         cmd_filter,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         TAKE12,
         ""
     },
     
     {
         "SecFilterDebugLevel",
         cmd_filter_debug_level,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         TAKE1,
         ""
     },

     {
         "SecFilterSelective",
         cmd_filter_selective,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         TAKE23,
         ""
     },

     {
         "SecFilterEngine",
         cmd_filter_engine,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         FLAG,
         ""
     },

     {
         "SecFilterScanPOST",
         cmd_scan_post,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         FLAG,
         ""
     },

     {
         "SecFilterDefaultAction",
         cmd_default_action,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         TAKE1,
         ""
     },

     {
         "SecAuditEngine",
         cmd_audit_engine,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         FLAG,
         ""
     },

     {
         "SecAuditLog",
         cmd_audit_log,
         NULL,
         RSRC_CONF | OR_AUTHCFG,
         TAKE1,
         ""
     },

    {NULL}
};

static int sec_logger (request_rec * r) {

#define AUDIT_BUF_SIZE 16384

    char *o = ap_palloc(r->pool, AUDIT_BUF_SIZE);
    sec_dir_config *dcfg = (sec_dir_config *)ap_get_module_config(r->per_dir_config, &security_module);
    array_header *arr;
    table_entry *te;
    char *remote_user, *local_user, *t;
    int i = 0;

    r = find_last_request(r);

    // ap_log_rerror(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, r, "mod_security: Audit log: [%s]", dcfg->auditlog_name); 

    // return immediatelly if we don't have a file to write to
    if (dcfg->auditlog_fd == -1) {
    	ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Audit log not found.", r->uri);
        return OK;
    }
    
    // ignore non-dynamic requests
    if (is_request_dynamic(r) == 0) return OK;

    strcpy (o, "");
    strncat (o, "========================================\n", AUDIT_BUF_SIZE);

    if (r->connection->remote_logname == NULL) {
        remote_user = "-";
    }
    else {
        remote_user = r->connection->remote_logname;
    }

    if (r->connection->user == NULL) {
        local_user = "-";
    }
    else {
        local_user = r->connection->user;
    }

    // todo: use r->request_time
    t = ap_psprintf (r->pool, "Request: %s %s %s [%s] \"%s\" %i %i\n",
		     r->connection->remote_ip, remote_user, local_user,
		     ap_get_time (), r->the_request, r->status,
		     r->bytes_sent);
	     
    strncat(o, t, AUDIT_BUF_SIZE);

    t = ap_psprintf (r->pool, "Handler: %s\n", r->handler);
    strncat(o, t, AUDIT_BUF_SIZE);

    strncat(o, "----------------------------------------\n", AUDIT_BUF_SIZE);

    // input headers
    t = ap_psprintf (r->pool, "%s\n", r->the_request);
    strncat(o, t, AUDIT_BUF_SIZE);

    arr = ap_table_elts(r->headers_in);

    te = (table_entry *) arr->elts;
    for (i = 0; i < arr->nelts; i++) {
	    t = ap_psprintf (r->pool, "%s: %s\n", te[i].key, te[i].val);
	    strncat (o, t, AUDIT_BUF_SIZE);
    }

    if (r->method_number == M_POST) {
    	// POST payload
	    request_body *rb = (request_body *) ap_table_get (r->notes, "mod_security-note");
    	if (rb != NULL) {
	        strncat (o, "\n", AUDIT_BUF_SIZE);
	        strncat (o, rb->buffer, AUDIT_BUF_SIZE);
    	    strncat (o, "\n", AUDIT_BUF_SIZE);
	    }
    	else {
	        strncat (o, "\n[POST payload not available]\n", AUDIT_BUF_SIZE);
    	}
    }

    strncat (o, "\n", AUDIT_BUF_SIZE);

    // output headers

    // the status first
    t = ap_psprintf(r->pool, "%s %s\n", r->protocol, r->status_line);
    strncat(o, t, AUDIT_BUF_SIZE);


    // output headers
    arr = ap_table_elts (r->headers_out);

    te = (table_entry *) arr->elts;
    for (i = 0; i < arr->nelts; i++) {
        t = ap_psprintf (r->pool, "%s: %s\n", te[i].key, te[i].val);
        strncat (o, t, AUDIT_BUF_SIZE);
    }

    // todo: output page

    // strncat(o, "========================================\n", AUDIT_BUF_SIZE);

    write (dcfg->auditlog_fd, o, strlen (o));

    return OK;
}

static int sec_check_access (request_rec * r) {
    sec_dir_config *dcfg = (sec_dir_config *) ap_get_module_config (r->per_dir_config, &security_module);

    // refuse to work if the per dir config is null
    if (dcfg == NULL) {
    	if (dcfg->filter_debug_level) {
		    // ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Filtering off, dcfg is null.");
    	}		
	    return DECLINED;
    }

    // refuse to work if this is a subrequest
    if (r->main != NULL) {
    	if (dcfg->filter_debug_level) {
	    	// ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Filtering off for a subrequest.");
    	}		
	    return DECLINED;
    }

    // refuse to work if the filtering is turned off
    if (dcfg->filter_engine != 1) {
    	if (dcfg->filter_debug_level) {    
	    	// ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Filtering off, switched off for path [%s]", dcfg->path);
    	}
	    return DECLINED;
    }

    if (dcfg->filter_debug_level) {
        ap_log_rerror (APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r, "mod_security: Checking with per-dir-config [%s][%s]", dcfg->path, r->uri);
    }

    return sec_check_all_signatures (r);
}

int is_request_dynamic (request_rec * r) {

    // find the last request                
    while (r->next != NULL) r = r->next;

    if (r->handler == NULL) {
    	return 0;
    }
    else {
        return 1;
    }
}

module MODULE_VAR_EXPORT security_module = {

    STANDARD_MODULE_STUFF,
    NULL,       			/* module initializer                  */
    sec_create_dir_config,	/* create per-dir    config structures */
    sec_merge_dir_config,	/* merge  per-dir    config structures */
    sec_create_srv_config,	/* create per-server config structures */
    NULL,			        /* merge  per-server config structures */
    sec_cmds,	    		/* table of config file commands       */
    NULL,			        /* [#8] MIME-typed-dispatched handlers */
    NULL,			        /* [#1] URI to filename translation    */
    NULL,			        /* [#4] validate user id from request  */
    NULL,			        /* [#5] check if the user is ok _here_ */
    sec_check_access,		/* [#3] check access by host address   */
    NULL,			        /* [#6] determine MIME type            */
    NULL,			        /* [#7] pre-run fixups                 */
    sec_logger,			    /* [#9] log a transaction              */
    NULL,			        /* [#2] header parser                  */
    NULL,			        /* child_init                          */
    NULL,			        /* child_exit                          */
    NULL			        /* [#0] post read-request              */
};
